@@ -25,7 +25,7 @@
   routine which basically goes through all possible paths in the TNFA
   and chooses the one which results in the best (leftmost and longest)
   match.  This can be spectacularly expensive and may run out of stack
-  space, but there really is no better known generic algorithm.  Quoting
+  space, but there really is no better known generic algorithm.	 Quoting
   Henry Spencer from comp.compilers:
   <URL: http://compilers.iecc.com/comparch/article/93-03-102>
 
@@ -47,7 +47,7 @@
 #endif /* HAVE_CONFIG_H */
 
 #ifdef TRE_USE_ALLOCA
-/* AIX requires this to be the first thing in the file.  */
+/* AIX requires this to be the first thing in the file.	 */
 #ifndef __GNUC__
 # if HAVE_ALLOCA_H
 #  include <alloca.h>
@@ -124,13 +124,13 @@ typedef struct tre_backtrack_struct {
 
 
 #ifdef TRE_USE_ALLOCA
-#define tre_bt_mem_new            tre_mem_newa
-#define tre_bt_mem_alloc          tre_mem_alloca
-#define tre_bt_mem_destroy(obj)   do { } while (0)
+#define tre_bt_mem_new		  tre_mem_newa
+#define tre_bt_mem_alloc	  tre_mem_alloca
+#define tre_bt_mem_destroy(obj)	  do { } while (0)
 #else /* !TRE_USE_ALLOCA */
-#define tre_bt_mem_new            tre_mem_new
-#define tre_bt_mem_alloc          tre_mem_alloc
-#define tre_bt_mem_destroy        tre_mem_destroy
+#define tre_bt_mem_new		  tre_mem_new
+#define tre_bt_mem_alloc	  tre_mem_alloc
+#define tre_bt_mem_destroy	  tre_mem_destroy
 #endif /* !TRE_USE_ALLOCA */
 
 
@@ -138,68 +138,68 @@ typedef struct tre_backtrack_struct {
   do									      \
     {									      \
       int i;								      \
-      if (!stack->next)					                      \
+      if (!stack->next)							      \
 	{								      \
-          tre_backtrack_t s;                                                  \
-	  s = tre_bt_mem_alloc(mem, sizeof(*s));	                      \
-	  if (!s)						              \
-            {                                                                 \
-	      tre_bt_mem_destroy(mem);                                        \
+	  tre_backtrack_t s;						      \
+	  s = tre_bt_mem_alloc(mem, sizeof(*s));			      \
+	  if (!s)							      \
+	    {								      \
+	      tre_bt_mem_destroy(mem);					      \
 	      if (tags)							      \
-                xfree(tags);						      \
+		xfree(tags);						      \
 	      if (pmatch)						      \
-                xfree(pmatch);						      \
+		xfree(pmatch);						      \
 	      if (states_seen)						      \
-                xfree(states_seen);					      \
+		xfree(states_seen);					      \
 	      return REG_ESPACE;					      \
-            }                                                                 \
-	  s->prev = stack;					              \
-	  s->next = NULL;					              \
-	  s->item.tags = tre_bt_mem_alloc(mem, 		                      \
-                                          sizeof(*tags) * tnfa->num_tags);    \
-	  if (!s->item.tags)	       			                      \
-            {                                                                 \
-	      tre_bt_mem_destroy(mem);                                        \
+	    }								      \
+	  s->prev = stack;						      \
+	  s->next = NULL;						      \
+	  s->item.tags = tre_bt_mem_alloc(mem,				      \
+					  sizeof(*tags) * tnfa->num_tags);    \
+	  if (!s->item.tags)						      \
+	    {								      \
+	      tre_bt_mem_destroy(mem);					      \
 	      if (tags)							      \
-                xfree(tags);						      \
+		xfree(tags);						      \
 	      if (pmatch)						      \
-                xfree(pmatch);						      \
+		xfree(pmatch);						      \
 	      if (states_seen)						      \
-                xfree(states_seen);					      \
+		xfree(states_seen);					      \
 	      return REG_ESPACE;					      \
-            }                                                                 \
-          stack->next = s;                                                    \
-          stack = s;                                                          \
-	}                                                                     \
-      else                                                                    \
+	    }								      \
+	  stack->next = s;						      \
+	  stack = s;							      \
+	}								      \
+      else								      \
 	stack = stack->next;						      \
       stack->item.pos = (_pos);						      \
-      stack->item.str_byte = (_str_byte);      				      \
-      BT_STACK_WIDE_IN;                                                       \
-      stack->item.state = (_state);    		 		              \
+      stack->item.str_byte = (_str_byte);				      \
+      BT_STACK_WIDE_IN;							      \
+      stack->item.state = (_state);					      \
       stack->item.state_id = (_state_id);				      \
       stack->item.next_c = (_next_c);					      \
       for (i = 0; i < tnfa->num_tags; i++)				      \
-	stack->item.tags[i] = (_tags)[i];      				      \
-      BT_STACK_MBSTATE_IN;                                                    \
+	stack->item.tags[i] = (_tags)[i];				      \
+      BT_STACK_MBSTATE_IN;						      \
     }									      \
   while (0)
 
-#define BT_STACK_POP()                                                        \
+#define BT_STACK_POP()							      \
   do									      \
-    {                                                                         \
+    {									      \
       int i;								      \
-      assert(stack->prev);                                                    \
+      assert(stack->prev);						      \
       pos = stack->item.pos;						      \
       str_byte = stack->item.str_byte;					      \
       BT_STACK_WIDE_OUT;						      \
       state = stack->item.state;					      \
       next_c = stack->item.next_c;					      \
       for (i = 0; i < tnfa->num_tags; i++)				      \
-        tags[i] = stack->item.tags[i];					      \
+	tags[i] = stack->item.tags[i];					      \
       BT_STACK_MBSTATE_OUT;						      \
       stack = stack->prev;						      \
-    }                                                                         \
+    }									      \
   while (0)
 
 #undef MIN
@@ -214,18 +214,17 @@ tre_tnfa_run_backtrack(const tre_tnfa_t *tnfa, const void *string,
   tre_char_t prev_c = 0, next_c = 0;
   const char *str_byte = string;
   int pos = 0;
-#ifdef TRE_MULTIBYTE
-  int pos_add_next = 1;
-#endif /* TRE_MULTIBYTE */
+  unsigned int pos_add_next = 1;
 #ifdef TRE_WCHAR
   const wchar_t *str_wide = string;
-#endif /* TRE_WCHAR */
 #ifdef TRE_MBSTATE
   mbstate_t mbstate;
 #endif /* TRE_MBSTATE */
+#endif /* TRE_WCHAR */
   int reg_notbol = eflags & REG_NOTBOL;
   int reg_noteol = eflags & REG_NOTEOL;
   int reg_newline = tnfa->cflags & REG_NEWLINE;
+  int str_user_end = 0;
 
   /* These are used to remember the necessary values of the above
      variables to return to the position where the current search
@@ -246,10 +245,10 @@ tre_tnfa_run_backtrack(const tre_tnfa_t *tnfa, const void *string,
   /* End offset of best match so far, or -1 if no match found yet. */
   int match_eo = -1;
   /* Tag arrays. */
-  int *next_tags, *tags;
+  int *next_tags, *tags = NULL;
   /* Current TNFA state. */
   tre_tnfa_transition_t *state;
-  int *states_seen;
+  int *states_seen = NULL;
 
   /* Memory allocator to for allocating the backtracking stack. */
   tre_mem_t mem = tre_bt_mem_new();
@@ -258,7 +257,8 @@ tre_tnfa_run_backtrack(const tre_tnfa_t *tnfa, const void *string,
   tre_backtrack_t stack;
 
   tre_tnfa_transition_t *trans_i;
-  regmatch_t *pmatch;
+  regmatch_t *pmatch = NULL;
+  int ret;
 
 #ifdef TRE_MBSTATE
   memset(&mbstate, '\0', sizeof(mbstate));
@@ -269,8 +269,8 @@ tre_tnfa_run_backtrack(const tre_tnfa_t *tnfa, const void *string,
   stack = tre_bt_mem_alloc(mem, sizeof(*stack));
   if (!stack)
     {
-      tre_bt_mem_destroy(mem);
-      return REG_ESPACE;
+      ret = REG_ESPACE;
+      goto error_exit;
     }
   stack->prev = NULL;
   stack->next = NULL;
@@ -284,8 +284,23 @@ tre_tnfa_run_backtrack(const tre_tnfa_t *tnfa, const void *string,
   states_seen = alloca(sizeof(*states_seen) * tnfa->num_states);
 #else /* !TRE_USE_ALLOCA */
   tags = xmalloc(sizeof(*tags) * tnfa->num_tags);
+  if (!tags)
+    {
+      ret = REG_ESPACE;
+      goto error_exit;
+    }
   pmatch = xmalloc(sizeof(*pmatch) * tnfa->num_submatches);
+  if (!pmatch)
+    {
+      ret = REG_ESPACE;
+      goto error_exit;
+    }
   states_seen = xmalloc(sizeof(*states_seen) * tnfa->num_states);
+  if (!states_seen)
+    {
+      ret = REG_ESPACE;
+      goto error_exit;
+    }
 #endif /* !TRE_USE_ALLOCA */
 
  retry:
@@ -303,6 +318,8 @@ tre_tnfa_run_backtrack(const tre_tnfa_t *tnfa, const void *string,
 
   state = NULL;
   pos = pos_start;
+  if (type == STR_USER)
+    str_source->rewind(pos + pos_add_next, str_source->context);
   GET_NEXT_WCHAR();
   pos_start = pos;
   next_c_start = next_c;
@@ -374,7 +391,7 @@ tre_tnfa_run_backtrack(const tre_tnfa_t *tnfa, const void *string,
 	    {
 	      int i;
 	      /* This match wins the previous match. */
-	      DPRINT(("  win previous\n"));
+	      DPRINT(("	 win previous\n"));
 	      match_eo = pos;
 	      if (match_tags)
 		for (i = 0; i < tnfa->num_tags; i++)
@@ -445,11 +462,14 @@ tre_tnfa_run_backtrack(const tre_tnfa_t *tnfa, const void *string,
 
 	  if (len < 0)
 	    {
+	      if (type == STR_USER)
+		result = str_source->compare(so, pos, bt_len,
+					     str_source->context);
 #ifdef TRE_WCHAR
-	      if (type == STR_WIDE)
+	      else if (type == STR_WIDE)
 		result = wcsncmp((wchar_t*)string + so, str_wide - 1, bt_len);
-	      else
 #endif /* TRE_WCHAR */
+	      else
 		result = strncmp((char*)string + so, str_byte - 1, bt_len);
 	    }
 	  else if (len - pos < bt_len)
@@ -474,34 +494,34 @@ tre_tnfa_run_backtrack(const tre_tnfa_t *tnfa, const void *string,
 
 	      /* Advance in input string and resync `prev_c', `next_c'
 		 and pos. */
-	      DPRINT(("  back reference matched\n"));
+	      DPRINT(("	 back reference matched\n"));
 	      str_byte += bt_len - 1;
 	      pos += bt_len - 1;
 	      GET_NEXT_WCHAR();
-	      DPRINT(("  pos now %d\n", pos));
+	      DPRINT(("	 pos now %d\n", pos));
 	    }
 	  else
 	    {
-	      DPRINT(("  back reference did not match\n"));
+	      DPRINT(("	 back reference did not match\n"));
 	      goto backtrack;
 	    }
 	}
       else
 	{
-	  /* Read the next character. */
-	  GET_NEXT_WCHAR();
-
 	  /* Check for end of string. */
 	  if (len < 0)
 	    {
-	      if (prev_c == L'\0')
+	      if (next_c == L'\0')
 		goto backtrack;
 	    }
 	  else
 	    {
-	      if (pos > len)
+	      if (pos >= len)
 		goto backtrack;
 	    }
+
+	  /* Read the next character. */
+	  GET_NEXT_WCHAR();
 	}
 
       next_state = NULL;
@@ -543,7 +563,7 @@ tre_tnfa_run_backtrack(const tre_tnfa_t *tnfa, const void *string,
 		}
 	      else
 		{
-		  /* Second mathing transition.  We may need to backtrack here
+		  /* Second mathing transition.	 We may need to backtrack here
 		     to take this transition instead of the first one, so we
 		     push this transition in the backtracking stack so we can
 		     jump back here if needed. */
@@ -580,7 +600,7 @@ tre_tnfa_run_backtrack(const tre_tnfa_t *tnfa, const void *string,
 	  /* A matching transition was not found.  Try to backtrack. */
 	  if (stack->prev)
 	    {
-	      DPRINT(("  backtracking\n"));
+	      DPRINT(("	 backtracking\n"));
 	      if (stack->item.state->assertions && ASSERT_BACKREF)
 		{
 		  DPRINT(("  states_seen[%d] = 0\n",
@@ -629,6 +649,10 @@ tre_tnfa_run_backtrack(const tre_tnfa_t *tnfa, const void *string,
 	}
     }
 
+  ret = match_eo >= 0 ? REG_OK : REG_NOMATCH;
+  *match_end_ofs = match_eo;
+
+ error_exit:
   tre_bt_mem_destroy(mem);
 #ifndef TRE_USE_ALLOCA
   if (tags)
@@ -639,6 +663,5 @@ tre_tnfa_run_backtrack(const tre_tnfa_t *tnfa, const void *string,
     xfree(states_seen);
 #endif /* !TRE_USE_ALLOCA */
 
-  *match_end_ofs = match_eo;
-  return match_eo >= 0 ? REG_OK : REG_NOMATCH;
+  return ret;
 }
